#include "graph.h"
#include "util.h"
#include "refine.h"

int SortFunc(const void *a,const void *b){

   if ((((sort_edge_type *)a)->wgt)>(((sort_edge_type *)b)->wgt))
     return -1;
   else if ((((sort_edge_type *)a)->wgt)<(((sort_edge_type *)b)->wgt))
     return 1;
   else 
     return 0;
}

void SortEdges(sort_edge_type *slist,int sedges){
 
  qsort(slist,(size_t)sedges,sizeof(sort_edge_type),SortFunc);
}


 
void AllocateGraph(graph_type *G){
  
  G->vlist=(vertex_type *)s_malloc((G->V+2)*sizeof(vertex_type));
  G->elist=(edge_type *)s_malloc((G->E)*sizeof(edge_type));
}

void FreeGraph(graph_type *G){
  
  free(G->vlist);
  free(G->elist);
}

int FindEdge(graph_type *G,int source,int dest){
  
  int i;								     

  for (i=G->vlist[source].start;i<G->vlist[source].end;++i)
    if (G->elist[i].adj==dest)
	  break;
  if (i<G->vlist[source].end)
    return i;
  else
    return -1;
}

void DeleteEdge(graph_type *G,int source,int dest){
  
  int i;

  i=FindEdge(G,source,dest);
  if (i!=-1){
    G->elist[i]=G->elist[G->vlist[source].end-1];
    --(G->vlist[source].end);
  }
  else
    Error_Exit("Edge Does not Exist");
}

void Read_TaskG(graph_type *G, FILE *ifp){
  int i,V;
  int eptr;
  vertex_type *vlist;
  edge_type *elist;
  char *sptr1,*sptr2,s[10000];
  
  eptr=0;
  G->vlist[1].start=0;
  V=G->V;
  vlist=G->vlist;
  elist=G->elist;
  for(i=1; i<=V;++i){
    do{
      fgets(s,10000,ifp);
	  if (strlen(s)==10000)
	    Error_Exit("Input string buffer overflow");
	}while (s[0]=='%');
	vlist[i].start=eptr;
	sptr1=s;
	for (;;){
	  elist[eptr].adj=strtol(sptr1,&sptr2,10);
	  if (sptr2==sptr1)
		break;
      elist[eptr].wgt=strtol(sptr2,&sptr1,10);
	  ++eptr;
	}
	vlist[i].end=eptr;
  }
  if (eptr!=G->E){
    Error_Exit("The graph has not proper quality");
  }   
}    	 

void Read_ExecT(int nproc,int ntask,int *ExecT,FILE *ifp){
  
  int i;

  for (i=1;i<=nproc*ntask;++i) 
	fscanf(ifp,"%d",(ExecT+i));
  fscanf(ifp,"\n");
}

void Read_Data(pint *ExecT,graph_type *TaskG,int *nproc,int *optval,char *filename){
  
  FILE *ifp;

  ifp = s_fopen(filename, "r");
  fscanf(ifp,"%d",nproc);
  fscanf(ifp,"%d",&(TaskG->V));
  fscanf(ifp,"%d",&(TaskG->E));
  //fscanf(ifp,"%d",optval);
  AllocateGraph(TaskG);
  *ExecT=(int *)s_malloc(sizeof(int)*(*nproc*TaskG->V+1));
  Read_ExecT(*nproc,TaskG->V,*ExecT,ifp);
  Read_TaskG(TaskG,ifp);
  fclose(ifp);
}

int FindQuality(int *ExecT,graph_type *TaskG,int *Assign){
  int i,m,V;
  int quality=0;
  vertex_type *vlist;
  edge_type *elist;

  V=TaskG->V;
  vlist=TaskG->vlist;
  elist=TaskG->elist;


  for (i=1;i<=V;++i){
    if (Assign[i]<=0)
	  Error_Exit("Assignment Not Comlete");
    for (m=vlist[i].start;m<vlist[i].end;++m){ 
	  if (Assign[i]!=Assign[elist[m].adj])
	    quality+=elist[m].wgt;
	}
  }
  quality=(int)(quality/2);
  for (i=1;i<=V;++i)
    quality+=ExecT[(Assign[i]-1)*V+i];
  return quality;
}

void Trivial(int *ExecT,graph_type *TaskG,int nproc,int *Assign){

  int i,m,temp;
  vertex_type *vlist;
  edge_type *elist;
  int V,min,minp;

  V=TaskG->V;
  vlist=TaskG->vlist;
  elist=TaskG->elist;

  min=-1;
  for (i=1;i<=nproc;++i){
    temp=0;
    for (m=1;m<=V;++m)
      temp+=ExecT[(i-1)*V+m];
    if ((min<0)||(temp<min)){
	  minp=i;
      min=temp;
	}
  }  
  for (i=1;i<=V;++i)
    Assign[i]=minp;
  printf("Trivial All to One Before Refine: %d\n",FindQuality(ExecT,TaskG,Assign));
  Refine(ExecT,TaskG,Assign,nproc);
  printf("Trivial All to One After Refine: %d\n",FindQuality(ExecT,TaskG,Assign));
  for (i=1;i<=V;++i){
    min=ExecT[i];
	Assign[i]=1;
	for (m=2;m<=nproc;++m)
	  if (min>ExecT[(m-1)*V+i]){
	    Assign[i]=m;
		min=ExecT[(m-1)*V+i];
	  }
  }
}

int BestProc(int *ExecT,int u,int nproc,int V){
 
   int min,minwgt,i;

   min=1;
   minwgt=ExecT[u];
   for (i=2;i<=nproc;++i){
     if (ExecT[(i-1)*V+u]<minwgt){
	   min=i;
	   minwgt=ExecT[(i-1)*V+u];
	 }
   }

   return min;
}

int TogetherBestProc(int *ExecT,int u,int v,int nproc,int V){
 
   int min,minwgt,temp,i;

   min=1;
   minwgt=ExecT[u]+ExecT[v];
   for (i=2;i<=nproc;++i){
     temp=ExecT[(i-1)*V+u]+ExecT[(i-1)*V+v];
     if (temp<minwgt){
	   min=i;
	   minwgt=temp;
	 }
   }

   return min;
}

int BestExecT(int *ExecT,int u,int nproc,int V){
 
   int minwgt,i;

   minwgt=ExecT[u];
   for (i=2;i<=nproc;++i){
     if (ExecT[(i-1)*V+u]<minwgt){
	   minwgt=ExecT[(i-1)*V+u];
	 }
   }

   return minwgt;
}
