#include "graph.h"
#include "util.h"
#include "mincut.h"
#include "vml.h"
#include "lamari.h"
#include "mylamari.h"
#include "procedge.h"
#include "refine.h"
#include "mlevel.h"
#include "twinass.h"
#include "multiple.h"
#include "timer.h"

void Convert_Option(char *arg,int *option);
void AssignTasks(int *ExecT,graph_type *TaskG,int nproc,int option,int *Cluster,int *Fixed,int *Assign,int *CAssign,int c_option);
void RefineAssign(int *ExecT,graph_type *TaskG,int nproc,int option,int *Cluster,int *Fixed,int *Assign);
void solveGraph(char* inputFileName, int option, int c_option,  long int* dbgTotalComp, long int* dbgTotalComm, int* beforeRefine, int* afterRefine, FILE* afp);
  
void Convert_Option(char *arg,int *option){

  if (arg!=NULL)  
    *option=atoi(arg);
  else
    *option=1; /*Default value */
}
void AssignTasks(int *ExecT,graph_type *TaskG,int nproc,int option,int *Cluster,int *Fixed,int *Assign,int *CAssign,int c_option){

  switch (option){
    case	0:
       Trivial(ExecT,TaskG,nproc,Assign);
	   break;		
    case	1:
       Vml(ExecT,TaskG,nproc,Fixed,Assign);
	   break;		
    case	2:
       Lamari(ExecT,TaskG,nproc,Assign);
	   break;
    case	3:
       MyLamari(ExecT,TaskG,nproc,Assign);
	   break;
	case	4:
       ProcEdge(ExecT,TaskG,nproc,Cluster,Assign,c_option);
	   break;
	case	5:
       MlevelAssign(ExecT,TaskG,nproc,Assign,CAssign,c_option);
	   break;
	case	6:
       TwinAssign(ExecT,TaskG,nproc,Cluster,Assign);
	   break;	
	case	7:
       MultiAssign(ExecT,TaskG,nproc,Assign);
	   break;	
	default :
	   Error_Exit("Invalid Parameter");
	   break;
   }
}

void RefineAssign(int *ExecT,graph_type *TaskG,int nproc,int option,int *Cluster,int *Fixed,int *Assign){

  switch (option){
  case	1:
    VmlRefine(ExecT,TaskG,Fixed,Assign,nproc);
    break;		
  case	4:
    TwoLevelRefine(ExecT,TaskG,Cluster,Assign,nproc);
    break;
  case	6:
    TwoLevelRefine(ExecT,TaskG,Cluster,Assign,nproc);
    break;	
  default :
    Refine(ExecT,TaskG,Assign,nproc);
    break;
   }
}

void solveGraph(char* inputFileName, int option, int c_option,  long int* dbgTotalComp, long int* dbgTotalComm, 
                int* beforeRefine, int* afterRefine, FILE* afp)
{
  graph_type TaskG;
  int *ExecT;
  int *CAssign,*Assign,*Cluster,*Fixed;
  int nproc,optval;
  int i;
timer boraTime; 
  Read_Data(&ExecT,&TaskG,&nproc,&optval,inputFileName);
   
   *dbgTotalComp = 0;
   for(i = 1; i <= TaskG.V * nproc; i++)
     *dbgTotalComp += ExecT[i];
   
   *dbgTotalComm = 0;
   for(i = 0; i < TaskG.vlist[TaskG.V].end; i++)
     *dbgTotalComm += TaskG.elist[i].wgt;
  
  Assign=(int *)s_malloc(sizeof(int)*(TaskG.V+1));
  CAssign=(int *)s_malloc(sizeof(int)*(TaskG.V+1));
  Fixed=(int *)s_malloc(sizeof(int)*(TaskG.V+1));
  Cluster=(int *)s_malloc(sizeof(int)*(TaskG.V+1));

   u_cleartimer(&boraTime);
   u_starttimer(&boraTime);
   
   AssignTasks(ExecT,&TaskG,nproc,option,Cluster,Fixed,Assign,CAssign,c_option);
   
   u_stoptimer(&boraTime);   
   u_printtimer(&boraTime, " BeforeRefineTime");

   fprintf(afp, "BeforeRefine\n");
  for(i = 1; i <= TaskG.V; i++)
     fprintf(afp, "%d ", Assign[i]);
   
  if (option==5) 
     *beforeRefine = FindQuality(ExecT,&TaskG,CAssign);
  else
    *beforeRefine = FindQuality(ExecT,&TaskG,Assign);

   u_cleartimer(&boraTime);
   u_starttimer(&boraTime);  
  
   RefineAssign(ExecT,&TaskG,nproc,option,Cluster,Fixed,Assign);
   u_stoptimer(&boraTime);   
   u_printtimer(&boraTime, " FMRefineTime");
   printf("\n");

   fprintf(afp, "\nAfterRefine\n");
   for(i = 1; i <= TaskG.V; i++)
     fprintf(afp, "%d ", Assign[i]);
   
   *afterRefine = FindQuality(ExecT,&TaskG,Assign);
  
  free(TaskG.vlist);
  free(TaskG.elist);
  free(Assign);
  free(CAssign);
  free(Fixed);
  free(Cluster);
  free(ExecT);
}   
int main(int argc,char *argv[])
{
  int option;
  int c_option;		
  int graphTypes;   
  graph_type TaskG;
 
  int beforeRefine;
  int afterRefine;
 
  char inputFileName[100];
  char outputFileName[100];
  char assignmentFileName[100];
  char debugFileName[100]; 
   
  int outputNo;
  
  long int dbgTotalComp;
  long int dbgTotalComm;

  char temp[10];

  char* names[8]; 
  char* PENames[3];
  char* MLNames[7];

  int opc;
  int c_opc;
    
  FILE *ofp;
  FILE *dfp;
  FILE *afp; 
   
  int extras[8];
  int i, j, k ,l, m, no;

  for(i = 0; i < 8; i++)
    names[i] = (char*)malloc(30*sizeof(char));
  for(i = 0; i < 3; i++)
    PENames[i] = (char*)malloc(30*sizeof(char));
  for(i = 0; i < 7; i++)
    MLNames[i] = (char*)malloc(30*sizeof(char));

  extras[1] = 1;  extras[2] = 1; extras[3] = 1;  extras[4] = 2;
  extras[5] = 6;  extras[6] = 1; extras[7] = 1;

  names[1] = "VML";
  names[2] = "Lamari" ;
  names[3] = "MyLamari"; 
  names[4] = "ProcEdge";
  names[5] = "Multilevel"; 
  names[6] = "TwinAssign";
  names[7] = "MultiAssign"; 
  
  PENames[1] = "Default";
  PENames[2] = "Grab";  

  MLNames[1] = "ProcEdgeNoAssign";
  MLNames[2] = "ProcEdgeWithAssign";
  MLNames[3] = "RandomCluster";
  MLNames[4] = "AgglCluster";
  MLNames[5] = "FullAgglCluster";
  MLNames[6] = "MultipleCluster";

   char heuristicName[100];
   
   int ProcType[10];
   double RCom[10];
   int RandomVertexType[50];
   int TreeVertexType[50];
   
   int noOfProcType = 0;
   int noOfRCom = 0;
   int noOfRandomVertexType = 0;
   int noOfTreeVertexType = 0;
   int noOfRandom;
   
   int randomVertexType, treeVertexType, procType;
   double rCom;
   
   char tchar[1000];
   int tint1, tint2;
   FILE *config = s_fopen("../configuration", "r");
   
   fscanf(config,"%s", tchar);
   tint1 = tint2 = 0;
   while(tint1 != -1)
     {
	fscanf(config,"%d %d", &tint1, &tint2);
	if(tint2 == 1)
	  {	     
	     RandomVertexType[noOfRandomVertexType] = tint1;
	                  noOfRandomVertexType++;
	  }
	
     }   
   
   fscanf(config,"%s", tchar);
   tint1 = tint2 = 0;
   while(tint1 != -1)
     {
	fscanf(config,"%d %d", &tint1, &tint2);
	if(tint2 == 1)
	  {
	     TreeVertexType[noOfTreeVertexType] = tint1;
	     noOfTreeVertexType++;
	  }
     }
   
   fscanf(config,"%s", tchar);
   tint1 = tint2 = 0;
   while(tint1 != -1)
     {
	fscanf(config,"%d %d", &tint1, &tint2);
	if(tint2 == 1)
	  {
	     
	     ProcType[noOfProcType] = tint1;
	     noOfProcType++;
	  }
     }
  
   fscanf(config,"%s", tchar);
   double tdouble1 = 0, tdouble2 = 0;
   while(tdouble1 != -1)
     {
	fscanf(config,"%lf %lf", &tdouble1, &tdouble2);
	if(tdouble2 == 1)
	  {
	     
	     RCom[noOfRCom] = tdouble1;
	     noOfRCom++;
	  }
     }
      
   fscanf(config,"%s", tchar);
   fscanf(config,"%d", &noOfRandom);
   
  if (argc>=4)
  {
    usRandomize;
    Convert_Option(argv[3],&option);
    Convert_Option(argv[4],&c_option);
    Convert_Option(argv[1],&graphTypes);
    Convert_Option(argv[2],&outputNo);    
     
    for(opc = 1; opc < 8; opc++)
    {
      for(c_opc = 1; c_opc <= extras[opc]; c_opc++)
      {
	if(option != 10)
	{
	  opc = option;
          c_opc = c_option;
        }

	if(graphTypes == 0 || graphTypes == 2)
	{
          strcpy(outputFileName, "Results/Excel/");
	  strcpy(debugFileName, "Results/Debug/");
	 	   
	  strcpy(tchar, names[opc]);
          
          if(option == 4 || (option == 10 && opc == 4))
	     strcat(tchar, PENames[c_opc]);
	  else if(option == 5 || (option == 10 && opc == 5))
	     strcat(tchar, MLNames[c_opc]);
	     
          strcat(tchar, "RandomGraph_");
          sprintf(temp,"No%d", outputNo);

	  strcpy(heuristicName, tchar); 
	   
	  strcat(outputFileName, tchar);
	  strcat(debugFileName, tchar);
	   
	  strcat(outputFileName, temp);
	  strcat(debugFileName, temp);	   
	   
          ofp = s_fopen(outputFileName, "w");
	  dfp = s_fopen(debugFileName, "w");
	   
	  for(i = 0; i < noOfRandomVertexType; i++)
	  {
	     randomVertexType = RandomVertexType[i];
	     for(j = 0; j < noOfProcType; j++)
	       {
		  procType = ProcType[j];
		  for(k = 0; k < noOfRCom; k++)
		    {
		       rCom = RCom[k];
		       for(l = 0; l < 4; l++)
			 {
			    for(m = 1; m <= noOfRandom; m++)
			      {
				 strcpy(inputFileName, "../dataGen/Data/");
				 strcpy(assignmentFileName, "Results/Assign/RandomGraph/");
				 
				 sprintf(temp,"dwt%d/", randomVertexType);
				 strcat(inputFileName,temp);
				  
				 strcat(assignmentFileName, temp);
				 strcat(assignmentFileName, names[opc]);
				 
				 if(option == 4 || (option == 10 && opc == 4))
				   strcat(assignmentFileName, PENames[c_opc]);
				 else if(option == 5 || (option == 10 && opc == 5))
				   strcat(assignmentFileName, MLNames[c_opc]);
				 strcat(assignmentFileName, "_");
			       
				 sprintf(temp,"t%d", randomVertexType);
				 strcat(inputFileName,temp);
				 strcat(assignmentFileName,temp);
				 
				 sprintf(temp,"p%d",  procType);
				 strcat(inputFileName,temp);
				 strcat(assignmentFileName,temp);
				 
				 sprintf(temp,"r%.1f", rCom);
				 strcat(inputFileName,temp);
				 strcat(assignmentFileName,temp);
				 
				 sprintf(temp,"ETC%d", l);
				 strcat(inputFileName,temp);
				 strcat(assignmentFileName,temp);
				 
				 sprintf(temp,".%d", m);
				 strcat(inputFileName,temp);
				 strcat(assignmentFileName,temp);
				 
				 sprintf(temp, "_No%d", outputNo);
				 strcat(assignmentFileName, temp);
				 				 
				 printf("Solving %s ",assignmentFileName);
				 
				 afp = s_fopen(assignmentFileName, "w");
				 solveGraph(inputFileName, opc, c_opc, &dbgTotalComp, &dbgTotalComm, &beforeRefine, &afterRefine, afp);
				 fclose(afp);
				 
				 if(beforeRefine < 0 || afterRefine < 0) 
				   {
				      printf("Bu nasil hata BR-AR\n");
				      fflush(0);
				      exit(12);
				   }
				 
				 fprintf(dfp, "%s\tComp-%ld\tComc-%ld\t BeforeRefine: %d\tAfterRefine: %d\n",inputFileName,dbgTotalComp, dbgTotalComm,
					 beforeRefine, afterRefine);

				 fprintf(ofp, "%d\t\t%d\n", beforeRefine, afterRefine);
			      }
			 }
		    }
	       }
	  }

	  fclose(ofp);
	  fclose(dfp);
	}
    
	if(graphTypes == 1 || graphTypes == 2)
	{
	   strcpy(outputFileName, "Results/Excel/");
	   strcpy(debugFileName, "Results/Debug/");
	   
	   strcpy(tchar, names[opc]);
	   
	   if(option == 4 || (option == 10 && opc == 4))
	     strcat(tchar, PENames[c_opc]);
	   else if(option == 5 || (option == 10 && opc == 5))
	     strcat(tchar, MLNames[c_opc]);
	   
	   strcat(tchar, "TreeGraph_");	   
	   sprintf(temp,"No%d", outputNo);
	   	   
	   strcpy(heuristicName, tchar);  
	   strcat(outputFileName, tchar);
	   strcat(debugFileName, tchar);
	   
	   strcat(outputFileName, temp);
	   strcat(debugFileName, temp);
	   	   
	   ofp = s_fopen(outputFileName, "w");
	   dfp = s_fopen(debugFileName, "w");
	   
	   for(i = 0; i < noOfTreeVertexType; i++) 
	     {
		treeVertexType = TreeVertexType[i];
		
		for(j = 0; j < noOfProcType; j++)
		  {
		    procType = ProcType[j];
		     
		       for(k = 0; k < noOfRCom; k++)
		       {
			  rCom = RCom[k];
			  for(l = 0; l < 4; l++)
			    {
			       for(no = 1; no <= noOfRandom; no++)
				 { 
				    strcpy(inputFileName, "../treeDataGen/Data/");
			            strcpy(assignmentFileName, "Results/Assign/TreeGraph/");
				    
				    sprintf(temp,"T%d/", treeVertexType);
				    strcat(inputFileName,temp);
				    				    
				    strcat(assignmentFileName, temp);
				    strcat(assignmentFileName, names[opc]);
				    
				    if(option == 4 || (option == 10 && opc == 4))
				      strcat(assignmentFileName, PENames[c_opc]);
				    else if(option == 5 || (option == 10 && opc == 5))
				      strcat(assignmentFileName, MLNames[c_opc]);
				    strcat(assignmentFileName, "_");				    
				    
				    sprintf(temp,"t%d",  treeVertexType);
				    strcat(inputFileName,temp);
				    strcat(assignmentFileName,temp);
				    
				    sprintf(temp,"p%d", procType);
				    strcat(inputFileName,temp);
				    strcat(assignmentFileName,temp);
				    
				    sprintf(temp,"r%.1f", rCom);
				    strcat(inputFileName,temp);
				    strcat(assignmentFileName,temp);
				    
				    sprintf(temp,"ETC%d", l);
				    strcat(inputFileName,temp);
				    strcat(assignmentFileName,temp);
				    
				    sprintf(temp,".%d", no);
				    strcat(inputFileName,temp);
				    strcat(assignmentFileName,temp);
				   
				    sprintf(temp, "_No%d", outputNo);
				    strcat(assignmentFileName, temp);
				    
				    printf("Solving %s ", assignmentFileName);
				    
				    afp = s_fopen(assignmentFileName, "w");
				    solveGraph(inputFileName, opc, c_opc, &dbgTotalComp, &dbgTotalComm, &beforeRefine, &afterRefine, afp);
				    fclose(afp);
				  
				    if(beforeRefine < 0 || afterRefine < 0) 
				      {
					 printf("Bu nasil hata BR-AR\n");
					 fflush(0);
					 exit(12);
				      }
				    
				    fprintf(dfp, "%s\tComp-%ld\tComc-%ld\t BeforeRefine: %d\tAfterRefine: %d\n",inputFileName,dbgTotalComp, dbgTotalComm,
					    beforeRefine, afterRefine);

				    fprintf(ofp, "%d\t\t%d\n", beforeRefine, afterRefine);
				 }
			    }
		       }
		  }
	     }
	   fclose(ofp);
	   fclose(dfp);
	}
	 
	 if(option != 10)
	   break;
      }
      if(option != 10)
	break;
    }
  }
   else{
      printf("Usage  : %s graphTypes outpuNo Option extraOption\n",argv[0]);
      printf("graphTypes <int>: 0 - all graphs, 1 - all trees, 2 - all\n\n");
      
      printf("Option :\n");
      printf(" 1   : VML\n");
      printf(" 2   : Lamari\n");
      printf(" 3   : MyLamari\n");
      printf(" 4   : ProcEdge\n");
      printf("      1 : Default[AC2]\n");
      printf("      2 : Grab\n");
      printf(" 5   : Multilevel\n");
      printf("    ExtraOption\n");
      printf("      1 : ProcEdge No Assign[ML-MC]\n");
      printf("      2 : ProcEdge With Assign\n");
      printf("      3 : Random Cluster[ML-RSAC]\n");
      printf("      4 : Aggl. Cluster[ML-SAC]\n");
      printf("      5 : Full Aggl. Cluster[ML-AC]\n");
      printf("      6 : Multiple Cluster[ML-MAC]\n");
      printf(" 6   : TwinAssign\n");
      printf(" 7   : MultiAssign\n");
      printf("10   : All\n");
      
      exit(1);
   }
   
   fclose(config);
   
   return 1;
}   
