
/*
  /-------------------------------------------------------------\
  |                     File : myheap.c                         |
  +-------------------------------------------------------------+
  |           Written by : Murat IKINCI                         |
  +-------------------------------------------------------------+
  |           Date       : 03.11.97                             |
  +-------------------------------------------------------------+
  |  Description : Heap                                         |
  +-------------------------------------------------------------+
  |  Modification by :                                          |
  |  Date            :                                          |
  |  Description     :                                          |
  \-------------------------------------------------------------/
*/
#include "util.h"
#include "myheap.h"

void Lock(pheap h,int id){
  h->ptr[id]=-1;
  --(h->numlocked);
  h->data[h->numlocked]=id;
}


void HeapAlloc(pheap h, int maxid){

  h->maxid = maxid;
  h->data =(int *) s_malloc(sizeof(int)*(h->maxid+1));
  h->ptr =(int *) s_malloc(sizeof(int)*(h->maxid+1));
  memset(h->ptr,0,(h->maxid+1)*sizeof(int));
  h->vals = (HeapValType *)  s_malloc(sizeof(HeapValType)*(h->maxid+1));
  h->size = 0;
  h->numlocked=maxid+1;
}
void HeapDealloc(pheap h){

  free(h->data);
  free(h->ptr);
  free(h->vals);
}
void HeapDeleteAll(pheap h){
  int i;

  if (h->size<h->maxid/3){
    for (i=1; i<=h->size; i++){
	  h->ptr[h->data[i]] = 0;
    }
  }
  else
    memset(h->ptr, 0 , (h->maxid+1)*sizeof(int));

  h->size = 0;
}


void HeapAdd(pheap h, int id){

  h->size++;
  h->data[h->size] = id;
  h->ptr[id] = h->size;
}

void MaxHeapUpdateKey(pheap h,int id,int newval){
  
  if (h->vals[id]>newval){
    h->vals[id]=newval;
    MaxHeapDecreaseKey(h,id);
  }
  else if (h->vals[id]<newval){
    h->vals[id]=newval;
    MaxHeapIncreaseKey(h,id);
  }
}

void MaxHeapInsert(pheap h, int id){
  register  int	i, j;


  ++(h->size);
  j = h->size;
  i = j >> 1;
  while (i>0 && h->vals[h->data[i]] < h->vals[id]){
    h->data[j] = h->data[i];
    h->ptr[h->data[i]] = j;
    j = i;
    i = i >> 1;
  }
  h->data[j] = id;
  h->ptr[id] = j;
}

void MaxHeapVerify(pheap h){
  register int 	i,j;
  int correct=1;


  i=1;
  j = i << 1;
  while (i<=(h->size>>1)){
    if (j<h->size && h->vals[h->data[j]] < h->vals[h->data[j+1]])
	  ++j;
    if (h->vals[h->data[j]] > h->vals[h->data[i]]){
	  correct=0;
	  i=h->size;
	}
    else{
	  ++i;
	  j = i << 1;
	}
  }
  if (!correct)
    printf("\nHeap Test Wrong");
  
}

void MaxHeapify(pheap h, int i){
  register int 	j;
  int root=h->data[i];

  j = i << 1;
  while (j<=h->size){
    if (j<h->size && h->vals[h->data[j]] < h->vals[h->data[j+1]])
	  ++j;
    if (h->vals[h->data[j]] <= h->vals[root])
	  break;
    else{
	  h->data[i] = h->data[j];
	  h->ptr[h->data[i]] = i;
	  i=j;
	  j = j << 1;
	}
  }
  h->data[i] = root;
  h->ptr[root] = i;
}


void MaxBuildHeap(pheap h)
{
int	i;

for (i=(h->size>>1); i>0; --i)
  MaxHeapify(h, i);
}


int MaxHeapExtractFirst(pheap h){
  if (h->size>1){
    int	ret=h->data[1], d=h->data[h->size];

    h->data[1] = h->data[h->size];
    h->ptr[h->data[h->size]] = 1;
    --(h->size);
    MaxHeapify(h, 1);
    h->ptr[ret] = 0;
    return ret;
  }
  else if (h->size==1){
    int	ret=h->data[1];
    h->size = 0;
    h->ptr[ret] = 0;
    return ret;    
  }
  else
    return -1;   /* nothing in heap*/
}


void MaxHeapIncreaseKey(pheap h,int id){
  int i=h->ptr[id];
  
  while ((i>>1)>0 && (h->vals[id]>=h->vals[h->data[i>>1]])){
    h->data[i]=h->data[i>>1];
	h->ptr[h->data[i]]=i;
    i=i>>1;
  }
  h->data[i]=id;
  h->ptr[id]=i;
}

void MaxHeapDelete(pheap h, int id){

  register int i=h->ptr[id];

  h->data[i] = h->data[h->size];
  h->ptr[h->data[i]] = i;
  --(h->size);
  h->ptr[id] = 0;
  if (i<=h->size){
    if (i>1&&h->vals[h->data[i]]>h->vals[h->data[i>>1]])
      MaxHeapIncreaseKey(h,h->data[i]);
    else
      MaxHeapDecreaseKey(h,h->data[i]);
  }
}

void Delete(pheap h,int id){
  
  h->data[h->ptr[id]]=h->data[h->size];
  h->ptr[h->data[h->size]]=h->ptr[id];
  h->ptr[id]=0;
  --(h->size);
}


int FindMaxHeapMin(pheap h){
  
  int i,m,min;

  min=h->data[1];
  for (i=(h->size/2)+1;i<=h->size;++i){
    m=h->data[i];
	if (h->vals[m]<h->vals[min])
	  min=m;
  }

  return min;
}


