#include "graph.h"
#include "util.h"
#include "mincut.h"

void FindReverseEdges(graph_type *TmG,int V,int *RevE){
  
  edge_type *elist;
  vertex_type *vlist;
  int i,m,u,k;

  vlist=TmG->vlist;
  elist=TmG->elist;
  for (i=1;i<=V;++i){
    for (m=vlist[i].start;m<vlist[i].end;++m){
	  u=elist[m].adj;
	  if (u>i){ 
		k=FindEdge(TmG,u,i);
		RevE[k]=m;
		RevE[m]=k;
      }
	}
  }
}

void GreedyCluster(int *ExecT,graph_type *TaskG,int nproc,int V,int source,double avgw,int *Assign){

  int *Fifo,fstart=0,fend=1;
  int i,m,mine,minp,temp;
  vertex_type *vlist;
  edge_type *elist;

  vlist=TaskG->vlist;
  elist=TaskG->elist;

  Fifo=(int *)s_malloc(sizeof(int)*(TaskG->V));
  Fifo[0]=source;
  Assign[source]=-1;
  while (fend>fstart){
    m=Fifo[fstart++];
	for (i=vlist[m].start;i<vlist[m].end;++i)
	  if ((elist[i].wgt>avgw)&&(Assign[elist[i].adj]==0)){
	    Assign[elist[i].adj]=-1;
		Fifo[fend++]=elist[i].adj;
	  }
  }

  minp=0;
  for (m=1;m<=nproc;++m){
    temp=0;
    for (i=0;i<fend;++i){
      temp+=ExecT[(m-1)*V+Fifo[i]];
    }
    if ((temp<mine)||(minp==0)){
      minp=m;
      mine=temp;
    }
  }
  for (i=0;i<fend;++i)
    Assign[Fifo[i]]=minp;
  free(Fifo);
}


void CreateTmG(graph_type *TaskG,int *Assign,graph_type *TmG){

  int eptr=0,i,m;
  edge_type *elist,e;
  vertex_type *vlist;

  elist=TmG->elist;
  vlist=TmG->vlist;
  TmG->V=0;

  for (i=1;i<=TaskG->V;++i){
    if (Assign[i]==0){
	  ++(TmG->V);
	  vlist[i].start=eptr;
      for (m=TaskG->vlist[i].start;m<TaskG->vlist[i].end;++m){    
	    e=TaskG->elist[m];
		if (Assign[e.adj]==0)
		  elist[eptr++]=e;
	  }
	  vlist[i].end=eptr;
	}
	else
	  vlist[i].start=vlist[i].end=eptr;
  }
  TmG->E=eptr;
}

void CreateSubGraph(int *ExecT,graph_type *TaskG,int *Assign,int nproc,int *TmW,graph_type *TmG,int *SumW){
  
  int eptr;
  edge_type e;
  int i,m,k,V,min,temp;

  edge_type *elist;
  vertex_type *vlist;
  int *TmX;
  int *SumX;

  eptr=0;
  V=TaskG->V;
  vlist=TmG->vlist;
  elist=TmG->elist;
  SumX=(int *)s_malloc(sizeof(int)*(V+1));
  TmX=(int *)s_malloc(sizeof(int)*(V*nproc+1));
  memcpy(TmX,ExecT,sizeof(int)*(V*nproc+1));

  for (i=1;i<=V;++i){
    if (Assign[i]==0){
      vlist[i].start=eptr;
	  elist[eptr++].adj=V+1;
	  elist[eptr++].adj=V+2;
	  for (m=TaskG->vlist[i].start;m<TaskG->vlist[i].end;++m){
	    e=TaskG->elist[m];
	    if (Assign[e.adj]==0){
	      elist[eptr].adj=e.adj;
          elist[eptr++].wgt=e.wgt*(nproc-1);
		}
		else{
		  for (k=1;k<=nproc;++k)
		    if (k!=Assign[e.adj])
		      TmX[(k-1)*V+i]+=e.wgt;
		}
	  }
	  vlist[i].end=eptr;
	}
	else
	  vlist[i].start=vlist[i].end=eptr;
  }
  vlist[V+1].start=eptr;
  for (i=1;i<=V;++i)
    if (Assign[i]==0)
      elist[eptr++].adj=i;
  vlist[V+1].end=eptr;
  vlist[V+2].start=eptr;
  for (i=1;i<=V;++i)
    if (Assign[i]==0)  
      elist[eptr++].adj=i;
  vlist[V+2].end=eptr;
  TmG->E=eptr;
  
  
  for (i=1;i<=V;++i)
    if (Assign[i]==0){
      SumX[i]=0;
	  for (m=1;m<=nproc;++m)
	    SumX[i]+=TmX[(m-1)*V+i];
	}

  for (i=1;i<=V;++i)
    if (Assign[i]==0){
	  min=0;
	  for (m=1;m<=nproc;++m){
	    temp=TmW[(m-1)*V+i]=SumX[i]-(nproc-1)*TmX[(m-1)*V+i];
		if (min>temp)
		  min=temp;
	  }
	  if (min<0){
	    for (m=1;m<=nproc;++m)
	      TmW[(m-1)*V+i]+=1-min;
	  }
	}
  for (i=1;i<=V;++i)
    if (Assign[i]==0){
	  SumW[i]=0;
	  for (m=1;m<=nproc;++m)
	    SumW[i]+=TmW[(m-1)*V+i];
	}

  free(TmX);
  free(SumX);
}


void CreateResG(int *SumW,int *TmW,graph_type *TmG,int q,int nproc,graph_type *ResG){
  
  int V,i,m;
  vertex_type *vlist;
  edge_type *elist;

  V=TmG->V-2;
  ResG->V=TmG->V;
  ResG->E=TmG->E;
  elist=ResG->elist;
  vlist=ResG->vlist;
  memcpy(elist,TmG->elist,sizeof(edge_type)*(TmG->E));
  memcpy(vlist,TmG->vlist,sizeof(vertex_type)*(TmG->V+1));
  /* Source Edges */
  for (i=vlist[V+1].start;i<vlist[V+1].end;++i){
    m=elist[i].adj;
	elist[i].wgt=TmW[(q-1)*V+m];
	elist[vlist[m].start].wgt=0;
  }
  /* Destination Edges */
  for (i=vlist[V+2].start;i<vlist[V+2].end;++i){
     m=elist[i].adj;
	 elist[i].wgt=0;
	 elist[vlist[m].start+1].wgt=SumW[m]-TmW[(q-1)*V+m];
  }
}

void Vml(int *ExecT,graph_type *TaskG,int nproc,int *Fixed,int *Assign){
  
  graph_type ResG,TmG;
  int V,q,flowval,newassigment=1;
  int *SumW;
  int *TmW;
  int *RevE;
  int min,minc,mine,i,m,temp,k,myn;
  double avgw;

  V=TaskG->V;
  TmG.V=V+2;
  TmG.E=TaskG->E+4*V;
  AllocateGraph(&TmG);

  ResG.V=TmG.V;
  ResG.E=TmG.E;
  AllocateGraph(&ResG);
  

  TmW=(int *)s_malloc(sizeof(int)*(V*nproc+1));
  SumW=(int *)s_malloc(sizeof(int)*(V+1));
  RevE=(int *)s_malloc(sizeof(int)*(TaskG->E+4*V+1));
  memset(Assign,0,sizeof(int)*(V+1));
  memset(Fixed,0,sizeof(int)*(V+1));
  while (newassigment){
    newassigment=0;
    CreateSubGraph(ExecT,TaskG,Assign,nproc,TmW,&TmG,SumW);
	FindReverseEdges(&TmG,V+2,RevE);
	for (q=1;q<=nproc;++q){
	  CreateResG(SumW,TmW,&TmG,q,nproc,&ResG);
	  MaxFlow(&ResG,RevE,V+1,V+2,&flowval);
	  BfsAssign(&ResG,V+1,q,Assign,&newassigment);
	}
  }

  /* Starting Lump */
  
  CreateTmG(TaskG,Assign,&TmG);
  FindReverseEdges(&TmG,V,RevE);
  if (TmG.V>0){
    if (TmG.V==1)
      minc=0;
    else{
      for (i=1;i<=V;++i)
	    if (Assign[i]==0)
	      break;
	  for (m=(i+1);m<=V;++m)
	    if (Assign[m]==0)
		  break;
	  ResG.V=V;
      ResG.E=TmG.E;
      memcpy(ResG.elist,TmG.elist,sizeof(edge_type)*(TmG.E));
      memcpy(ResG.vlist,TmG.vlist,sizeof(vertex_type)*(V+2));
	  MaxFlow(&ResG,RevE,i,m,&minc);
	  for(m=m+1;m<=V;++m)
	    if (Assign[m]==0){
	      memcpy(ResG.elist,TmG.elist,sizeof(edge_type)*(TmG.E));
	      MaxFlow(&ResG,RevE,i,m,&flowval);
		  if (flowval<minc)
		    minc=flowval;
		}
	}
	for (i=1;i<=V;++i)
	  if (Assign[i]==0){	
	    temp=ExecT[i];
	    for (m=2;m<=nproc;++m){
	      if (ExecT[(m-1)*V+i]<temp)
	        temp=ExecT[(m-1)*V+i];
		}
	    minc+=temp;
	  } 	
	/* minc includes the L value for VML Lump */
	min=0;
	for (i=1;i<=nproc;++i){
	  temp=0;
	  for (m=1;m<=V;++m){
	    if (Assign[m]==0){
	      temp+=ExecT[(i-1)*V+m];
		  for (k=TaskG->vlist[m].start;k<TaskG->vlist[m].end;++k){
		    myn=TaskG->elist[k].adj;
			if ((Assign[myn]!=0)&&(Assign[myn]!=i))
			  temp+=TaskG->elist[k].wgt;
		  }
		}
	  }
	  if ((temp<mine)||(min==0)){
	    min=i;
        mine=temp;
      }
	}
	if (mine<=minc){
	  TmG.V=0;
	  for (i=1;i<=V;++i)
	    if (Assign[i]==0)
	      Assign[i]=min;
	}
  }	 		  
  /* Finding Fixed tasks */
  memcpy(Fixed,Assign,sizeof(int)*(V+1));
  /* Starting Greedy*/
  if (TmG.V>0){
    avgw=0;
    for (i=1;i<=V;++i){
      for (m=TaskG->vlist[i].start;m<TaskG->vlist[i].end;++m)
	      avgw+=TaskG->elist[m].wgt;
    }
    avgw=avgw/(V*(V-1));	  
    for (i=1;i<=V;++i)
      if (Assign[i]==0)
	    GreedyCluster(ExecT,&TmG,nproc,V,i,avgw,Assign);
  }                      

  free(RevE);
  free(SumW);
  free(TmW);
  FreeGraph(&ResG);
  FreeGraph(&TmG);
}
