#include "graph.h"
#include "util.h"
#include <time.h>

int* rank;
int* px;

typedef struct node{
	int u,v;
} node;

int Arrive(graph_type *G,int source,int dest){
	
	int *Fifo,fstart=0,fend=1;
	int i,m;
	int *Color;
	vertex_type *vlist;
	edge_type *elist;
	
	vlist=G->vlist;
	elist=G->elist;
	
	Fifo=(int *)s_malloc(sizeof(int)*G->V);
	Color=(int *)s_malloc(sizeof(int)*(G->V+1));
	memset(Color,0,sizeof(int)*(G->V+1));
	Fifo[0]=source;
	Color[source]=1;
	while ((fend>fstart)&&(Color[dest]==0)){
		m=Fifo[fstart++];
		for (i=vlist[m].start;i<vlist[m].end;++i)
			if (Color[elist[i].adj]==0){
				Color[elist[i].adj]=1;
				Fifo[fend++]=elist[i].adj;
			}
	}
	m=Color[dest];
	free(Color);
	free(Fifo);
	return m;
}

void buLink(int x, int y)
{
	if(rank[x] > rank[y])
		px[y] = x;
	else
	{	
		px[x] = y;
		if(rank[x] == rank[y])
			rank[y]= rank[y]+1;
	}
	return;
}


int buFindSet(int s)
{
	
	if(s != px[s])
		px[s] = buFindSet(px[s]);
	return px[s];
}

void buUnion(int x, int y)
{
	
	buLink(buFindSet(x), buFindSet(y));
	
	return ;
}


int B_Arrive(graph_type *G,int source,int dest)
{
	int ps = buFindSet(source);
	int pd = buFindSet(dest);
	if(pd == ps)
		return 1;
	else
		return 0;
}

int Random(int m){
	return (int)(rand()%m);
}

void Permute(node *List,int V){
	int i,m,ptr;
	node anode;
	
	ptr=1;
	for (i=1;i<V;++i)
		for (m=i+1;m<=V;++m){
			List[ptr].u=i;
			List[ptr++].v=m;
		}
	for (i=(V*(V+1)/2);i>1;--i){
		m=Random(i)+1;
		anode=List[i];
		List[i]=List[m];
		List[m]=anode;
	}
}

void program(int notask, char* outputFileName){
	
	graph_type TaskG;
	int *Parent;
	node *List;
	int i,m,ptr,u,v,optr;
	
	int buEdgCnt = 0;
	TaskG.V=notask;
	TaskG.E=TaskG.V-1;
	TaskG.vlist=(vertex_type *)s_malloc(sizeof(vertex_type)*(TaskG.V+1));
	TaskG.elist=(edge_type *)s_malloc(sizeof(edge_type)*(TaskG.V*TaskG.V));
	
	Parent=(int *)s_malloc(sizeof(int)*(TaskG.V+1));
	List=(node *)s_malloc(sizeof(node)*(1+(TaskG.V+1)*TaskG.V/2));
	
	Permute(List,TaskG.V);
	
	memset(Parent,0,sizeof(int)*(TaskG.V+1));
	Parent[1]=-1;
	for (i=1;i<=TaskG.V;++i)
		TaskG.vlist[i].start=TaskG.vlist[i].end=(i-1)*TaskG.V;
	
	for (i=1;i<=TaskG.V*(TaskG.V+1)/2;++i){
		u=List[i].u;
		v=List[i].v;
		if (B_Arrive(&TaskG,u,v))
			continue;
		
		if ((Parent[v]==0)||(Parent[u]==0))
		{
			buUnion(u,v);
			buEdgCnt ++;
			TaskG.elist[TaskG.vlist[u].end++].adj=v;
			TaskG.elist[TaskG.vlist[v].end++].adj=u;
			if (Parent[v]==0)
				Parent[v]=u;
			else if (Parent[u]==0)
				Parent[u]=v;	   
		}
		if(buEdgCnt == 9999)
			break;
		
	}
	
	ptr=0;
	for (i=1;i<=TaskG.V;++i){
		optr=ptr;
		for (m=TaskG.vlist[i].start;m<TaskG.vlist[i].end;++m){
			if (TaskG.elist[m].adj>i)
				TaskG.elist[ptr++]=TaskG.elist[m];
		}
		TaskG.vlist[i].start=optr;
		TaskG.vlist[i].end=ptr;
	}
	Write_Data(&TaskG, outputFileName);
	free(TaskG.vlist);
	free(TaskG.elist);
	free(Parent);
	free(List);
}

int main(int argc,char *argv[])
{
	
	int i, j;
	char outputFileName[300]={'\0'};
	char temp[10];
	
	srand(time(NULL));
	
	
	int ProcType[10];
	double RCom[10];
	int VertexType[50];
	
	int noOfProcType = 0;
	int noOfRCom = 0;
	int noOfVertexType = 0;
	int noOfRandom = 20;
	
	int vertexType, procType;
	double rCom;
	
	char tchar[1000];
	int tint1, tint2;
	FILE *config = s_fopen("../configuration", "r");
	
	fscanf(config,"%s", tchar);
	tint1 = tint2 = 0;
	while(tint1 != -1)
	{	
		fscanf(config,"%d %d", &tint1, &tint2);
	}
	
	fscanf(config,"%s", tchar);
	tint1 = tint2 = 0;
	while(tint1 != -1)
	{
		
		fscanf(config,"%d %d", &tint1, &tint2);
		if(tint2 == 1){						
			VertexType[noOfVertexType] = tint1;
			noOfVertexType++;
		}
	}
	
	fscanf(config,"%s", tchar);
	tint1 = tint2 = 0;
	while(tint1 != -1)
	{
		
		fscanf(config,"%d %d", &tint1, &tint2);
		if(tint2 == 1) {						
			ProcType[noOfProcType] = tint1;
			noOfProcType++;
		}
	}
	
	fscanf(config,"%s", tchar);
	double tdouble1 = 0, tdouble2 = 0;
	while(tdouble1 != -1)
	{
		
		fscanf(config,"%lf %lf", &tdouble1, &tdouble2);
		if(tdouble2 == 1.0 ){
			
			RCom[noOfRCom] = tdouble1;
			noOfRCom++;
		}
	}
	fscanf(config,"%s", tchar);

	fscanf(config,"%d", &noOfRandom);
	srand((unsigned)time( NULL ));
	fclose(config);
	
	for(i = 0; i < noOfVertexType; i++)
	{
		int no;
		vertexType = VertexType[i];
		for(no = 1; no <= noOfRandom; no++)
		{
			
			int ll;
			strcpy(outputFileName, "Data/");
			
			sprintf(temp,"t%d", vertexType);
			strcat(outputFileName,temp);
			
			sprintf(temp,".%d", no);
			strcat(outputFileName,temp);
			
			printf("%s\n", outputFileName);
			rank = (int*) malloc(sizeof(int)*(vertexType+1));
			px = (int*) malloc(sizeof(int)*(vertexType+1));
			for(ll= 1; ll <= vertexType; ll++)
			{
				px[ll] = ll;
				rank[ll] = 0;
			}
			program(vertexType, outputFileName);
			free(rank);
			free(px);
		}
	}
	
	
	return 1;
}
